{-# LANGUAGE ExistentialQuantification,
             FlexibleInstances,
             MultiParamTypeClasses,
             ScopedTypeVariables,
             UndecidableInstances,
             FlexibleContexts
  #-}

module AOP.JPStackT (
 EJP(..),
 JPStackT,
 JPStack,
 MonadJPStack(..),
 pcCflow,
 pcCflowbelow,
 runJPStackT,
 jpStackAspect
) where

import AOP.Base hiding (run)
import AOP.Internal.StableNamesEq
import Debug.Trace

data EJP =  forall a b m. (Monad m, PolyTypeable (a -> m b)) => EJP (a -> m b) FunctionTag a
type JPStack = [EJP]
newtype JPStackT m a = JPStackT {run :: JPStack -> m (a, JPStack)}

runJPStackT :: Monad m => JPStackT m a -> JPStack -> m a
runJPStackT c l = do
               (v, _) <- run c l
               return v

instance Monad m => Monad (JPStackT m) where
   return k = JPStackT $ \s -> return (k, s)
   ma >>= f = JPStackT $ \s0 -> do
                               (value, s1) <- run ma s0
                               run (f value) s1

class Monad m => MonadJPStack m where
      getJPStack :: m JPStack
      addJpToStack :: EJP -> m ()
      popJPStack :: m ()

instance Monad m => MonadJPStack (JPStackT m) where
      getJPStack = JPStackT (\s -> return (s, s))
      addJpToStack jp = JPStackT (\s -> return ((), jp:s))
      popJPStack = JPStackT (\(jp:s) -> return ((), s))

{- Interaction with AOT -}
instance (MonadJPStack m, Typeable1Monad (AOT m)) => MonadJPStack (AOT m) where
         getJPStack = lift getJPStack
         addJpToStack jp = AOT $ \aenv -> do addJpToStack jp; return ((), aenv)
         popJPStack = lift popJPStack

{- Interaction with AOT_secure -}
instance (MonadJPStack m, Typeable1Monad (AOT_s m)) => MonadJPStack (AOT_s m) where
         getJPStack = lift getJPStack
         addJpToStack jp = AOT_s $ \aenv pc -> do addJpToStack jp; return ((), (aenv, pc))
         popJPStack = lift popJPStack

{- Interaction with AOT_secure_protected -}
instance (MonadJPStack m, Typeable1Monad (AOT_sp m)) => MonadJPStack (AOT_sp m) where
         getJPStack = lift getJPStack
         addJpToStack jp = AOT_sp $ \aenv pc -> do addJpToStack jp; return ((), (aenv, pc))
         popJPStack = lift popJPStack


{- Interaction with standard monad transformers -}
instance Typeable1Monad m => Typeable1 (JPStackT m) where
         typeOf1 _ = mkTyConApp (mkTyCon3 "JPStackT" "JPStackT" "JPStackT")
                     [typeOf1 (undefined :: m ())]

instance MonadTrans JPStackT where
         lift ma = JPStackT $ \s -> do
                           a <- ma
                           return (a, s)

instance MonadIO m => MonadIO (JPStackT m) where
         liftIO = lift . liftIO

instance Monad m => MonadState s (JPStackT (StateT s m)) where
         get = lift get
         put s = lift (put s)

instance MonadWriter w m => MonadWriter w (JPStackT m) where
    tell     = lift . tell
    listen m = JPStackT $ \ctx -> do
                              ((a, ctx'), w) <- listen (run m ctx)
                              return ((a, w), ctx')
    pass m   = JPStackT $ \ctx -> pass $ do
                                   ((a, f), ctx') <- run m ctx
                                   return ((a, ctx'), f)

instance MonadCont m => MonadCont (JPStackT m) where
    callCC f = JPStackT $ \aenv ->
        callCC $ \c ->
        run (f (\a -> JPStackT $ \aenv' -> c (a, aenv'))) aenv

instance MonadReader r m => MonadReader r (JPStackT m) where
    ask       = lift ask
    local f m = JPStackT $ \s -> local f (run m s)


-- | Comparing identity of functions:
-- | When given a join point with a regular function (signaled by the default tag)
-- | then we use StableNames for comparison. If the functions are wrapped,
-- | we compare the tags
compareFun' :: t -> FunctionTag -> EJP -> Bool
compareFun' f ft (EJP g t _) = if t == defaultFunctionTag
                                then stableNamesEq f g
                                else ft == t

-- | PcCflow/PcCflowbelow Pointcut Definition

_pcCflow f t = return (\jp -> do jpStack <- getJPStack
                                 return (any (compareFun' f t) jpStack))

_pcCflowbelow f t = return (\jp -> do jpStack <- getJPStack
                                      if null jpStack
                                         then return False
                                         else return (any (compareFun' f t) (tail jpStack)))

class PcCflow f where 
      pcCflow :: (Typeable1Monad m, MonadJPStack m, PolyTypeable (f a (m b))) =>
               f a (m b) -> PC m c (m' d)

instance PcCflow ((->)) where 
         pcCflow f = PC (_pcCflow f defaultFunctionTag)

instance PcCflow Function where
         pcCflow (Function f t) = PC (_pcCflow f t)

class PcCflowbelow f where 
      pcCflowbelow :: (Typeable1Monad m, MonadJPStack m, PolyTypeable (f a (m b))) =>
                    f a (m b) -> PC m c (m' d)

instance PcCflowbelow ((->)) where 
         pcCflowbelow f = PC (_pcCflowbelow f defaultFunctionTag)

instance PcCflowbelow Function where
         pcCflowbelow (Function f t) = PC (_pcCflowbelow f t)

-- | Jp collecting aspect
callAny :: (Typeable1Monad m, MonadJPStack m) => PC m a b
callAny = PC $ return (\(Jp f t a) -> do
                             addJpToStack (EJP f t a)
                             return True)

-- | Pushes the current jp to the join point stack
collectAdv proceed arg = do result <- proceed arg
                            popJPStack
                            return result

-- | Join points are collected by a regular aspect
-- | deployed at some bootstrapping time.
jpStackAspect :: (Typeable1Monad m, MonadJPStack m) => Aspect m c1 (m c2) c1 c2
jpStackAspect = aspect callAny collectAdv